/*
 * Copyright (C) 1995  Sam Moore, Warren Jasper
 * Copyright (C) 1997  Tim Edwards
 * Copyright (C) 1999  Igor Volobouev
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

/*----------------------------------------------------------------*/
/* ciodac02.c							  */
/*----------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <asm/io.h>

#if LINUX_VERSION_CODE >= 0x020200
#include <asm/uaccess.h>
#else
#include <asm/segment.h>
#endif

#include "dac02.h"

#if LINUX_VERSION_CODE >= 0x020100
#if LINUX_VERSION_CODE < 0x020200
#error Linux kernels 2.1.x are not supported. Please switch to a stable kernel.
#endif
#endif

#define BYTE unsigned char
#define LONG unsigned long

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

/* Default device major number */

#ifndef DEFAULT_MAJOR_DEV
#define DEFAULT_MAJOR_DEV  30    /* Default Major Device Number */
#endif

/* Base I/O Address */

#ifndef BASE_REG
#define BASE_REG         0x310   /* Default base address for CIO-DAC02/16 */
#endif

/* Number of channels */

#ifndef CHANNELS
#define CHANNELS	2
#endif

#define A_OFFSET	0
#define B_OFFSET	1
#define PORT_OFFSET	2

#define A_MASK		0xff
#define B_MASK		0xff00

typedef struct CHANNEL_REC {
   int           open;        /* for lockout of other processes */
   unsigned char mode;        /* for readback of 82C55 mode */
} ChanRec;

static ChanRec Chan[CHANNELS];              /* Channel specific information   */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in       */

#ifndef TRUE
#define TRUE  1
#endif
#ifndef FALSE
#define FALSE 0
#endif


/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

#if LINUX_VERSION_CODE >= 0x020200
static ssize_t dac02_write(struct file *filePtr, const char *buf, 
			   size_t count, loff_t *offset)
{
    int  minor = MINOR(filePtr->f_dentry->d_inode->i_rdev);
#else
static int dac02_write(struct inode *iNode, struct file *filePtr,
		       const char *buf, int count)
{	
    int  minor = MINOR(iNode->i_rdev);
#endif
    int  portbase = BASE_REG + (minor * PORT_OFFSET);
    long value;

    if ( minor >= CHANNELS ) {
        printk("dac02_write: bad minor number: %d\n", minor );
        return -1;
    }
#if LINUX_VERSION_CODE >= 0x020200
    get_user(value, (long *)buf);
#else
    value = get_fs_long(buf);
#endif
    outb_p( (BYTE) (value & A_MASK),        portbase + A_OFFSET);
    outb_p( (BYTE)((value & B_MASK) >> 8),  portbase + B_OFFSET); 

    #ifdef DEBUG
        printk("DAC02 set to %#x\n", value & 0xffff);
    #endif

    return 2;
}


/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int dac02_open(struct inode *iNode, struct file *filePtr)
{
    unsigned long flags;
    int minor = MINOR(iNode->i_rdev);

    /* 
       check if device is already open: only one process may write to a
       port at a time.  There is still the possibility of two processes 
       writing to two different channels messing things up. However,
       the overhead to check for this may not be worth it.
    */

    save_flags(flags);
    cli();
    if ( Chan[minor].open == TRUE ) {
	restore_flags(flags);
        return -EBUSY;
    } else {
	Chan[minor].open = TRUE;
	Chan[minor].mode = 0;
	restore_flags(flags);
    }

    #ifdef DEBUG
        printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, Chan[minor].mode);
    #endif
    return 0;   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

#if LINUX_VERSION_CODE >= 0x020200
static int dac02_close(struct inode *iNode, struct file *filePtr)
#else
static void dac02_close(struct inode *iNode, struct file *filePtr)
#endif
{
    int minor = MINOR(iNode->i_rdev);

    #ifdef DEBUG
        printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
    #endif

    Chan[minor].open = FALSE;
#if LINUX_VERSION_CODE >= 0x020200
    return 0;
#endif
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

static int dac02_ioctl(struct inode *iNode, struct file *filePtr,
		       unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  int portbase = BASE_REG + (minor * PORT_OFFSET);

  switch (cmd) {
    case DAC02_SETA:
      outb_p((BYTE)arg, portbase + A_OFFSET);
      break;

    case DAC02_SETB:
      outb_p((BYTE)arg, portbase + B_OFFSET);
      break;

    default:
      return(-EINVAL);
      break;
  }
  return 0;
}

/***************************************************************************
 *
 * See /usr/include/linux/fs.h  for more information
 *
 ***************************************************************************/

static struct file_operations dac02_fops = {
	write:      dac02_write,
	ioctl:      dac02_ioctl,
	open:       dac02_open,
	release:    dac02_close,
};

/***************************************************************************
 *
 * Loads driver. Called when "insmod dac02.o" is invoked on the command line.
 *
 ***************************************************************************/

int init_module(void) 
{
  int i;

  /* Register as a device with kernel.  */

  if (register_chrdev(MajorNumber, "dac02", &dac02_fops)) {
    printk("%s: Failure to load module.\n", ADAPTER_ID);
    return -EIO;
  } else {
    printk("%s: address=%#x.",ADAPTER_ID, BASE_REG);
    printk(" 06/05/99 igv@lbl.gov\n");
  }

  /* Set all channel structures to show nothing active/open */
  for (i = 0; i < CHANNELS; i++)
    Chan[i].open = FALSE;

  return 0;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod dac02" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
    if (MOD_IN_USE) {
      printk("%s: device busy, remove delayed\n", ADAPTER_ID);
    }

    if (unregister_chrdev(MajorNumber, "dac02") != 0) {
      printk("%s: cleanup_module failed\n", ADAPTER_ID);
    } else {

    #ifdef DEBUG
        printk("%s: module removed.\n", ADAPTER_ID);
    #endif

    }    
}
